/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "dchublistmanager.h"

#include <stdlib.h>

#include <QApplication>
#include <QObject>
#include <QToolButton>
#include <QLineEdit>
#include <QCursor>
#include <QTabWidget>
#include <QComboBox>
#include <QCheckBox>
#include <QMessageBox>
#include <QClipboard>
#include <QResizeEvent>
#include <QShowEvent>
#include <QEvent>
#include <QMdiArea>
#include <QMdiSubWindow>
#include <QHeaderView>

#include <vector>

#include "dcmenuhandler.h"
#include "dchubprofile.h"
#include "dchubfilter.h"
#include "dciconloader.h"
#include "dcconnectionmanager.h"
#include "dcconfig.h"
#include "dchublistitems.h"
#include "dcguiutils.h"

#include <dclib/dcos.h>
#include <dclib/core/cstringlist.h>
#include <dclib/dcobject.h>

#include "dceditserver.h"
#include "cdialogmessage.h"
#include "publichubsmodel.h"
#include "publichubsproxy.h"

DCHubListManager *DCHubListManager::pHubListManager = NULL;

/** */
DCHubListManager::DCHubListManager(QWidget * parent) : QWidget(parent) {
    setupUi(this);

    /* icons moved out of .ui files so that they can be themed */
    ToolButton_ADDFILTER->setIcon(QIcon(DCIconLoader::Instance()->GetPixmap(eiEDITADD)));
    ToolButton_EDITFILTER->setIcon(QIcon(DCIconLoader::Instance()->GetPixmap(eiEDIT)));
    ToolButton_DELFILTER->setIcon(QIcon(DCIconLoader::Instance()->GetPixmap(eiEDITDELETE)));
    ToolButton_RELOADPUBLICHUBLIST->setIcon(QIcon(DCIconLoader::Instance()->GetPixmap(eiRELOAD)));
    ToolButton_UPDATEPUBLICHUBLIST->setIcon(QIcon(DCIconLoader::Instance()->GetPixmap(eiUPDATE)));
    ToolButton_CONNECT->setIcon(QIcon(DCIconLoader::Instance()->GetPixmap(eiCONNECT)));

    /* ... and for bookmark buttons */
    ToolButton_BOOKMARK_SAVE->setIcon(QIcon(DCIconLoader::Instance()->GetPixmap(eiSAVE)));
    ToolButton_BOOKMARK_ADD->setIcon(QIcon(DCIconLoader::Instance()->GetPixmap(eiEDITADD)));
    ToolButton_BOOKMARK_EDIT->setIcon(QIcon(DCIconLoader::Instance()->GetPixmap(eiEDIT)));
    ToolButton_BOOKMARK_DEL->setIcon(QIcon(DCIconLoader::Instance()->GetPixmap(eiEDITDELETE)));

    lastPublicHubsWidth = -1;
    lastBookmarkHubsWidth = -1;

    /* we have custom column resizing code */
    TreeView_PUBLIC->header()->setStretchLastSection(false);
    treeView_BOOKMARKS->header()->setStretchLastSection(false);

    // set default icon
    setWindowIcon(DCIconLoader::Instance()->GetPixmap(eiSERVER));

    if ((parent != 0) && (qobject_cast<QMdiArea*>(parent) != 0)) {
        m_pContainerWindow = new QMdiSubWindow();
        m_pContainerWindow->setWidget(this);
    } else {
        m_pContainerWindow = 0;
    }

    model = new PublicHubsModel(this);
    proxy = new PublicHubsProxy(this);
    proxy->setSortRole(PublicHubsProxyRole);
    proxy->setSortCaseSensitivity(Qt::CaseInsensitive);
    proxy->setFilterRole(PublicHubsProxyRole);
    proxy->setFilterCaseSensitivity(Qt::CaseInsensitive);
    proxy->setSourceModel(model);
    TreeView_PUBLIC->setModel(proxy);

    m_pMessageList = new QList<CDCMessage*>();

    m_pHubListModel = NULL;

    InitDocument();
}

/** */
DCHubListManager::~DCHubListManager() {

    m_HLMMutex.lock();

    if (m_pMessageList) {
        QList<CDCMessage*> * tmp = m_pMessageList;
        m_pMessageList = 0;

        for (QList<CDCMessage*>::const_iterator it = tmp->constBegin(); it != tmp->constEnd(); ++it) {
            delete *it;
        }

        delete tmp;
    }

    m_HLMMutex.unlock();

    /* models deleted by QObject auto child deletion */

    // we might have removed public hubs
    if (DCConfig::Instance()->GetRemovePublicHubEnabled()) {
        DCConfig::Instance()->SaveDCPublicHub();
    }

    // bookmarks may have been sorted and the new order needs to be saved
    updateBookmarkOrder();

    /* delete all hub filter objects */
    for (QMap<QString, DC_HubFilterObject*>::const_iterator it = m_HubFilterMap.constBegin(); it != m_HubFilterMap.constEnd(); ++it) {
        delete it.value();
    }
    m_HubFilterMap.clear();

    if (m_pContainerWindow) {
        m_pContainerWindow->setWidget(0); // otherwise the QMdiSubWindow will delete this again
        delete m_pContainerWindow;
        m_pContainerWindow = 0;
    }
}

/** current tab widget change slot */
void DCHubListManager::slotTabWidgetCurrentChange(QWidget*) {
    SizeColumnsPreservingRatios();
}

/** overridden so that the column widths are initialized on first show() */
void DCHubListManager::showEvent(QShowEvent * event) {
    QWidget::showEvent(event);

    if (isVisible()) {
        SizeColumnsPreservingRatios();
    }
}

/** resize event handler */
void DCHubListManager::resizeEvent(QResizeEvent *) {
    SizeColumnsPreservingRatios();
}

/** */
void DCHubListManager::closeEvent(QCloseEvent * e) {
    emit onDie(this);
    QWidget::closeEvent(e);

    if (m_pContainerWindow && m_pContainerWindow->parent()) {
        /* Using QMdiArea::removeSubWindow() breaks tabs mode but this works */
        m_pContainerWindow->setParent(0);
    }
}

/** Initialize or adjust widths of the ListView columns */
void DCHubListManager::SizeColumnsPreservingRatios() {
    int width;

    if (TreeView_PUBLIC->isVisible()) {
        width = TreeView_PUBLIC->width();
        if (width > 0) {
            if (lastPublicHubsWidth == -1) {
                /* set initial widths using default ratios */
                /* also expand the list view since we have lots of columns now */
                TreeView_PUBLIC->setColumnWidth(0, ((width * 3) / 10));
                TreeView_PUBLIC->setColumnWidth(1, ((width * 3) / 10));
                TreeView_PUBLIC->setColumnWidth(2, ((width * 3) / 10));
                TreeView_PUBLIC->setColumnWidth(3, ((width * 1) / 10));
                TreeView_PUBLIC->setColumnWidth(4, ((width * 2) / 10));
                TreeView_PUBLIC->setColumnWidth(5, ((width * 1) / 10));
                TreeView_PUBLIC->setColumnWidth(6, ((width * 1) / 10));
                TreeView_PUBLIC->setColumnWidth(7, ((width * 3) / 10));

                lastPublicHubsWidth = TreeView_PUBLIC->width();
            } else if (lastPublicHubsWidth != width) {
                DCGuiUtils::AdjustColumnWidths(TreeView_PUBLIC, lastPublicHubsWidth);
                lastPublicHubsWidth = TreeView_PUBLIC->width();
            }
        }
    } else if (treeView_BOOKMARKS->isVisible()) {
        width = treeView_BOOKMARKS->width();
        if (width > 0) {
            if (lastBookmarkHubsWidth == -1) {
                treeView_BOOKMARKS->setColumnWidth(COLUMN_HUB_AUTOCONNECT, width / 7);
                treeView_BOOKMARKS->setColumnWidth(COLUMN_HUB_ADDRESS, width / 7);
                treeView_BOOKMARKS->setColumnWidth(COLUMN_HUB_DESC, width / 7);
                treeView_BOOKMARKS->setColumnWidth(COLUMN_HUB_ENCODING, width / 7);
                treeView_BOOKMARKS->setColumnWidth(COLUMN_HUB_NAME, width / 7);
                treeView_BOOKMARKS->setColumnWidth(COLUMN_HUB_NICK, width / 7);
                treeView_BOOKMARKS->setColumnWidth(COLUMN_HUB_PASSWORD, width / 7);

                lastBookmarkHubsWidth = treeView_BOOKMARKS->width();
            } else if (lastBookmarkHubsWidth != width) {
                DCGuiUtils::AdjustColumnWidths(treeView_BOOKMARKS, lastBookmarkHubsWidth);
                lastBookmarkHubsWidth = treeView_BOOKMARKS->width();
            }
        }
    }
}

/** */
void DCHubListManager::SetToolBar(QFrame * frame1, QFrame * frame2, bool vis) {
    if (vis) {
        frame1->hide();
        frame1->setEnabled(false);
        frame2->show();
        frame2->setEnabled(true);
    } else {
        frame2->hide();
        frame2->setEnabled(false);
        frame1->show();
        frame1->setEnabled(true);
    }
}

/** */
void DCHubListManager::InitDocument() {
    StringMap * map;
    bool btb1 = true, btb2 = true;
    bool bfilter = false;

    DC_HubFilterObject * HubFilterObject;

    // hide progressbar
    ProgressBar_LOADHUBLIST->setMinimum(0);
    ProgressBar_LOADHUBLIST->hide();

    // set bookmark icon
    TabWidget_SERVERLIST->setTabIcon(1, QIcon(DCIconLoader::Instance()->GetPixmap(eiBOOKMARK_FOLDER)));

    // init the toolbars
    Frame_TOOLBAR1_HIDE->installEventFilter(this);
    Frame_TOOLBAR1_SHOW->installEventFilter(this);
    Frame_TOOLBAR2_HIDE->installEventFilter(this);
    Frame_TOOLBAR2_SHOW->installEventFilter(this);

    checkBox_AUTOCONNECT->setChecked(DCConfig::Instance()->GetAutoConnectToAllHubs());

    m_pHubListModel = new HubListModel(this);
    treeView_BOOKMARKS->setModel(m_pHubListModel);

    // restore settings
    if (DCConfig::Instance()->GetMap("HUBVIEW", map)) {
        if (((*map)["WIDTH"].toInt() > 0) && ((*map)["HEIGHT"].toInt() > 0)) {
            if (m_pContainerWindow != 0) {
                m_pContainerWindow->setGeometry((*map)["X"].toInt(), (*map)["Y"].toInt(), (*map)["WIDTH"].toInt(), (*map)["HEIGHT"].toInt());
            }
        }

        if ((!((*map)["TOOLBAR1"].isEmpty())) && ((*map)["TOOLBAR1"].toInt() == 0)) {
            btb1 = false;
        }

        if ((!((*map)["TOOLBAR2"].isEmpty())) && ((*map)["TOOLBAR2"].toInt() == 0)) {
            btb2 = false;
        }

        // restore active page
        TabWidget_SERVERLIST->setCurrentIndex((*map)["ACTIVEPAGE"].toInt());
        // restore linedit
        LineEdit_SERVER->setText((*map)["LE_SERVER"]);

        /* restore sorting parameters */
        TreeView_PUBLIC->sortByColumn((*map)["PUBLICSORTCOLUMN"].toInt(), DCGuiUtils::SortOrderFromName((*map)["PUBLICSORTORDER"]));

        if ((!((*map)["BOOKMARKSORTINGENABLED"].isEmpty())) && ((*map)["BOOKMARKSORTINGENABLED"].toInt() == 0)) {
            treeView_BOOKMARKS->setSortingEnabled(false);
        } else {
            treeView_BOOKMARKS->setSortingEnabled(true);
        }

//        //treeView_BOOKMARKS->sort((*map)["BOOKMARKSORTCOLUMN"].toInt(), DCGuiUtils::SortOrderFromName((*map)["BOOKMARKSORTORDER"]));
    }

    SetToolBar(Frame_TOOLBAR1_HIDE, Frame_TOOLBAR1_SHOW, btb1);
    SetToolBar(Frame_TOOLBAR2_HIDE, Frame_TOOLBAR2_SHOW, btb2);

    connect(checkBox_AUTOCONNECT, SIGNAL(clicked()), this, SLOT(slotAutoConnectToAllHubs()));

    connect(LineEdit_SERVER, SIGNAL(returnPressed()), this, SLOT(slotConnect()));
    connect(ToolButton_RELOADPUBLICHUBLIST, SIGNAL(clicked()), this, SLOT(slotReloadPublicHubList()));
    connect(ToolButton_UPDATEPUBLICHUBLIST, SIGNAL(clicked()), this, SLOT(slotUpdatePublicHubList()));
    connect(ToolButton_CONNECT, SIGNAL(clicked()), this, SLOT(slotConnect()));

    // public list
    connect(TreeView_PUBLIC, SIGNAL(activated(const QModelIndex &)), this, SLOT(slotItemActivatedPublic(const QModelIndex &)));
    connect(TreeView_PUBLIC, SIGNAL(customContextMenuRequested(const QPoint &)), this, SLOT(slotContextMenuPublicList(const QPoint &)));

    // bookmark list
    connect(treeView_BOOKMARKS, SIGNAL(clicked(const QModelIndex&)), this, SLOT(slotItemClickedBookmarks(const QModelIndex&)));
    connect(treeView_BOOKMARKS, SIGNAL(customContextMenuRequested(const QPoint &)), this, SLOT(slotContextMenuBookmarkList(const QPoint &)));
    /*connect(treeView_BOOKMARKS, SIGNAL(itemChanged (QTreeWidgetItem*, int)), this, SLOT(slotItemChangedBookmarks(QTreeWidgetItem*,int)));*/
    connect(ToolButton_BOOKMARK_UP, SIGNAL(clicked()), this, SLOT(slotMoveBookmarkUp()));
    connect(ToolButton_BOOKMARK_DOWN, SIGNAL(clicked()), this, SLOT(slotMoveBookmarkDown()));
    //connect(ToolButton_BOOKMARK_ENABLE_SORT, SIGNAL(clicked()), this, SLOT(slotEnableBookmarkSorting()));
    connect(ToolButton_BOOKMARK_SAVE, SIGNAL(clicked()), this, SLOT(slotSaveBookmarks()));
    connect(ToolButton_BOOKMARK_ADD, SIGNAL(clicked()), this, SLOT(slotAddBookmark()));
    connect(ToolButton_BOOKMARK_EDIT, SIGNAL(clicked()), this, SLOT(slotEditBookmark()));
    connect(ToolButton_BOOKMARK_DEL, SIGNAL(clicked()), this, SLOT(slotDelBookmark()));

    connect(TabWidget_SERVERLIST, SIGNAL(currentChanged(QWidget*)), this, SLOT(slotTabWidgetCurrentChange(QWidget*)));

    // filter
    connect(ToolButton_ADDFILTER, SIGNAL(clicked()), this, SLOT(slotAddFilter()));
    connect(ToolButton_EDITFILTER, SIGNAL(clicked()), this, SLOT(slotEditFilter()));
    connect(ToolButton_DELFILTER, SIGNAL(clicked()), this, SLOT(slotDelFilter()));
    connect(ComboBox_FILTER, SIGNAL(activated(const QString &)), this, SLOT(slotFilterChange(const QString &)));

    treeView_BOOKMARKS->setAlternatingRowColors(DCConfig::Instance()->GetAlternatingRowColors());
    TreeView_PUBLIC->setAlternatingRowColors(DCConfig::Instance()->GetAlternatingRowColors());
    connect(DCConfig::Instance(), SIGNAL(alternatingRowColorsChanged(bool)), this, SLOT(slotAltRowColors(bool)));

    connect(&m_Timer, SIGNAL(timeout()), this, SLOT(timerDone()));

    // load filter
    DCConfig::Instance()->LoadDCHubFilter(&m_HubFilterMap);

    // create default
    if (m_HubFilterMap.isEmpty()) {
        HubFilterObject = new DC_HubFilterObject();

        HubFilterObject->m_sFilterName = "Default";

        m_HubFilterMap.insert(HubFilterObject->m_sFilterName, HubFilterObject);

        DCConfig::Instance()->SaveDCHubFilter(&m_HubFilterMap);
    }

    for (QMap<QString, DC_HubFilterObject*>::const_iterator it = m_HubFilterMap.constBegin(); it != m_HubFilterMap.constEnd(); ++it) {
        HubFilterObject = it.value();

        ComboBox_FILTER->addItem(HubFilterObject->m_sFilterName);

        if ((*map)["FILTER"] == HubFilterObject->m_sFilterName) {
            bfilter = true;
        }
    }

    // restore filter
    if (bfilter) {
        int index = ComboBox_FILTER->findText((*map)["FILTER"]);
        if (index != -1) {
            ComboBox_FILTER->setCurrentIndex(index);
        }
    } else {
        int index = ComboBox_FILTER->findText("Default");
        if (index != -1) {
            ComboBox_FILTER->setCurrentIndex(index);
        }
    }

    ShowBookmarkList();
    ShowPublicHubList();

    m_Timer.setSingleShot(true);
    m_Timer.start(500);
}

/** */
void DCHubListManager::DeInitDocument() {
    StringMap * map;

    // save hub view settings
    DCConfig::Instance()->GetMap("HUBVIEW", map);

    if (m_pContainerWindow != 0) {
        (*map)["X"] = QString().setNum(m_pContainerWindow->x());
        (*map)["Y"] = QString().setNum(m_pContainerWindow->y());
        (*map)["WIDTH"] = QString().setNum(m_pContainerWindow->width());
        (*map)["HEIGHT"] = QString().setNum(m_pContainerWindow->height());
        (*map)["VISIBLE"] = QString().setNum(m_pContainerWindow->isVisible());
        (*map)["MAXIMIZED"] = QString().setNum(m_pContainerWindow->isMaximized());
        (*map)["MINIMIZED"] = QString().setNum(m_pContainerWindow->isMinimized());
    }
    (*map)["ACTIVEPAGE"] = QString().setNum(TabWidget_SERVERLIST->currentIndex());
    (*map)["LE_SERVER"] = LineEdit_SERVER->text();
    (*map)["FILTER"] = ComboBox_FILTER->currentText();

    if (!isEnabled()) {
        setEnabled(true);
    }

    (*map)["TOOLBAR1"] = QString().setNum(Frame_TOOLBAR1_SHOW->isEnabled());
    (*map)["TOOLBAR2"] = QString().setNum(Frame_TOOLBAR2_SHOW->isEnabled());

    (*map)["PUBLICSORTCOLUMN"] = QString().setNum(TreeView_PUBLIC->header()->sortIndicatorSection());
    (*map)["PUBLICSORTORDER"] = DCGuiUtils::SortOrderName(TreeView_PUBLIC->header()->sortIndicatorOrder());

 //   (*map)["BOOKMARKSORTCOLUMN"] = QString().setNum(treeView_BOOKMARKS->sortColumn());
 //   (*map)["BOOKMARKSORTORDER"] = DCGuiUtils::SortOrderName(treeView_BOOKMARKS->header()->sortIndicatorOrder());
 //   (*map)["BOOKMARKSORTINGENABLED"] = QString().setNum(treeView_BOOKMARKS->isSortingEnabled());

    if (m_pHubListModel){
        m_pHubListModel->clearModel();

        delete m_pHubListModel;

        m_pHubListModel = NULL;
    }
}

/** event filter */
bool DCHubListManager::eventFilter(QObject* object, QEvent* event) {
    if (event->type() == QEvent::MouseButtonDblClick) {
        if (object == Frame_TOOLBAR1_HIDE) {
            SetToolBar(Frame_TOOLBAR1_HIDE, Frame_TOOLBAR1_SHOW, true);
        } else if (object == Frame_TOOLBAR1_SHOW) {
            SetToolBar(Frame_TOOLBAR1_HIDE, Frame_TOOLBAR1_SHOW, false);
        } else if (object == Frame_TOOLBAR2_HIDE) {
            SetToolBar(Frame_TOOLBAR2_HIDE, Frame_TOOLBAR2_SHOW, true);

            // show progressbar
            if (m_pHubListUrl != 0)
                ProgressBar_LOADHUBLIST->show();
        } else if (object == Frame_TOOLBAR2_SHOW) {
            SetToolBar(Frame_TOOLBAR2_HIDE, Frame_TOOLBAR2_SHOW, false);
        }
    }

    return QWidget::eventFilter(object, event); // standard event processing
}

/** callback function */
int DCHubListManager::DC_CallBack(CDCMessage * DCMessage) {
    int err = -1;

    m_HLMMutex.lock();

    if (DCMessage && m_pMessageList) {
        m_pMessageList->append(DCMessage);
        err = 0;
    }

    m_HLMMutex.unlock();

    return err;
}

/** */
void DCHubListManager::timerDone() {
    CDCMessage * DCMsg;
    int i;

    for (i = 0; i < 50; i++) {
        if (m_HLMMutex.tryLock() == false) {
            break;
        }

        if (m_pMessageList && !m_pMessageList->isEmpty()) {
            DCMsg = m_pMessageList->takeFirst();
        } else {
            DCMsg = 0;
        }

        m_HLMMutex.unlock();

        if (DCMsg == 0) {
            break;
        }

        switch (DCMsg->m_eType) {
            case DC_MESSAGE_GETHUBLIST:
            {
                DCMessageGetHubList * msg = (DCMessageGetHubList*) DCMsg;

                if (msg->m_bRun == false) {
                    // enable buttons
                    ToolButton_RELOADPUBLICHUBLIST->setEnabled(true);
                    ToolButton_UPDATEPUBLICHUBLIST->setEnabled(true);

                    // hide progressbar
                    ProgressBar_LOADHUBLIST->hide();

                    ShowPublicHubList();
                } else {
                    // disable buttons
                    ToolButton_RELOADPUBLICHUBLIST->setEnabled(false);
                    ToolButton_UPDATEPUBLICHUBLIST->setEnabled(false);

                    // show progressbar
                    ProgressBar_LOADHUBLIST->show();
                }

                break;
            }

            case DC_MESSAGE_TRANSFER:
            {
                CMessageTransfer *msg = (CMessageTransfer*) DCMsg;

                ProgressBar_LOADHUBLIST->setMaximum(msg->m_nLength);
                ProgressBar_LOADHUBLIST->setValue(msg->m_nTransfered);

                break;
            }

            default:
            {
                break;
            }
        }

        if (DCMsg) {
            delete DCMsg;
        }
    }

    m_Timer.setSingleShot(true);
    m_Timer.start(500);
}

/** */
void DCHubListManager::ShowPublicHubList() {
    DCConfigHubItem * hubitem = 0;
    CList<DCConfigHubItem> PublicHubList;
    int x = 0;

    // disable filter combo box
    ComboBox_FILTER->setEnabled(false);

    TabWidget_SERVERLIST->setTabText(0, tr("Public") + " (0/0)");

    TreeView_PUBLIC->setUpdatesEnabled(false);
    TreeView_PUBLIC->setSortingEnabled(false);
    TreeView_PUBLIC->clearSelection();
    model->clear();

    DCConfig::Instance()->GetPublicHubList(&PublicHubList);

    while ((hubitem = PublicHubList.Next(hubitem)) != 0) {
        ++x;

        if (x == 100) {
            qApp->processEvents();
            x = 0;
        }

        model->addHub(
                QString::fromAscii(hubitem->m_sName.Data()),
                QString::fromAscii(hubitem->m_sHost.Data()),
                QString::fromAscii(hubitem->m_sDescription.Data()),
                hubitem->m_nUserCount,
                QString::fromAscii(hubitem->m_sCountry.Data()),
                hubitem->m_nShared,
                hubitem->m_nMinShare,
                QString::fromAscii(hubitem->m_sExtra.Data())
                );
    }

    TreeView_PUBLIC->setSortingEnabled(true);
    proxy->invalidate();
    slotFilterChange(ComboBox_FILTER->currentText());
    TreeView_PUBLIC->setUpdatesEnabled(true);

    TabWidget_SERVERLIST->setTabText(0, tr("Public") + " (" + QString().setNum(TreeView_PUBLIC->model()->rowCount()) + "/" + QString().setNum(PublicHubList.Count()) + ")");

    // enable filter combo box
    ComboBox_FILTER->setEnabled(true);
}

/** */
void DCHubListManager::ShowBookmarkList() {
    DCConfigHubItem * hubitem = 0;
    CList<DCConfigHubItem> list;
    //bool sortingenabled = treeView_BOOKMARKS->isSortingEnabled();

    DCConfig::Instance()->GetBookmarkHubList(&list);

    treeView_BOOKMARKS->setSortingEnabled(false);
    treeView_BOOKMARKS->setUpdatesEnabled(false);
    m_pHubListModel->clearModel();;

    while ((hubitem = list.Next(hubitem)) != 0) {
        DCConfigHubProfile pConfigHubProfile;

        if (!DCConfig::Instance()->GetHubProfile(hubitem->m_sName, &pConfigHubProfile)){
            //Loading defaults into the profile
            pConfigHubProfile.m_sNick = "";
            pConfigHubProfile.m_sPassword = "";
            pConfigHubProfile.m_sRemoteEncoding = "";
            pConfigHubProfile.m_bAutoConnect = false;
            pConfigHubProfile.m_sName = hubitem->m_sName;
            pConfigHubProfile.m_bSSL           = false;
            pConfigHubProfile.m_bTag           = false;
            pConfigHubProfile.m_bExtHubCount   = true;

            // add new profile
            DCConfig::Instance()->AddHubProfile(&pConfigHubProfile);

            // save profile
            DCConfig::Instance()->SaveHubProfile();

            // set bookmark for that hub
            DCConfig::Instance()->SetBookmarkHubProfile(pConfigHubProfile.m_sName, pConfigHubProfile.m_sName);

            // update myinfo
            DCConnectionManager::Instance()->SendMyInfoToConnectedServers();
        }

        QString hub_name, hub_addr, hub_nick, hub_pass, hub_desc, hub_enc;
        bool hub_auto = false;
        QList<QVariant> data;

        hub_name = QString(hubitem->m_sName.Data());
        hub_addr = QString(hubitem->m_sHost.Data());
        hub_desc = QString(hubitem->m_sDescription.Data());
        hub_nick = QString(pConfigHubProfile.m_sNick.Data());
        hub_pass = QString(pConfigHubProfile.m_sPassword.Data());
        hub_enc  = QString(pConfigHubProfile.m_sRemoteEncoding.Data());
        hub_auto = pConfigHubProfile.m_bAutoConnect;

        if (hub_nick.isEmpty() || hub_nick.isNull())
            hub_nick = tr("Default nick");
        if (hub_enc.isEmpty()  || hub_enc.isNull())
            hub_enc  = tr("Default encoding");

        data << hub_auto << hub_name << hub_addr << hub_nick << hub_pass << hub_desc << hub_enc;

        m_pHubListModel->addResult(data);
    }

    treeView_BOOKMARKS->setUpdatesEnabled(true);

    TabWidget_SERVERLIST->setTabText(1, tr("Bookmarks") + " (" + QString().setNum(treeView_BOOKMARKS->model()->rowCount()) + ")");
}

/** */
void DCHubListManager::slotReloadPublicHubList() {
    DCConfig::Instance()->ClearPublicHubList();

    GetPublicHubList();
}

/** update public hub list */
void DCHubListManager::slotUpdatePublicHubList() {
    GetPublicHubList();
}

/** */
void DCHubListManager::slotItemActivatedPublic(const QModelIndex & index) {
    PublicHubItem * item = model->getItem(proxy->mapToSource(index));

    if (item)
        DCConnectionManager::Instance()->Connect(item->name.toAscii().constData(), item->server.toAscii().constData());
}

/** */
void DCHubListManager::slotItemClickedBookmarks(const QModelIndex &index) {
    if (!index.isValid())
        return;

    if (!index.column() == COLUMN_HUB_AUTOCONNECT)
        return;

    HubListModelItem *item = (HubListModelItem*)index.internalPointer();

    if (!item)
        return;

    item->updateColumn(COLUMN_HUB_AUTOCONNECT, !(bool)item->data(COLUMN_HUB_AUTOCONNECT).toBool());

    QString hub_name = item->data(COLUMN_HUB_NAME).toString();
    DCConfigHubProfile pConfigHubProfile;

    DCConfig::Instance()->GetHubProfile(hub_name.toAscii().constData(), &pConfigHubProfile);

    pConfigHubProfile.m_bAutoConnect = item->data(COLUMN_HUB_AUTOCONNECT).toBool();

    DCConfig::Instance()->AddHubProfile(&pConfigHubProfile);
    DCConfig::Instance()->SaveHubProfile();

    m_pHubListModel->repaint();
}

/** */
void DCHubListManager::slotConnect() {
    QString hub = LineEdit_SERVER->text();
    if (hub.mid(0,8) == "dchub://")
    {
        hub.replace(0,8,"");
    }
    if (hub != "") {
        DCConnectionManager::Instance()->Connect(CString(), hub.toAscii().constData());
        DCConfig::Instance()->AppendQuickConnectHistory(hub);
    }
}

/** */
void DCHubListManager::slotAddFilter() {
    DC_HubFilterObject * HubFilterObject;
    DCHubFilter * hf;

    hf = new DCHubFilter(this);

    if (hf->exec() == QDialog::Accepted) {
        HubFilterObject = new DC_HubFilterObject();

        // get the new data
        hf->Get(HubFilterObject);

        if (HubFilterObject->m_sFilterName.isEmpty()) {
            // todo: error message

            delete HubFilterObject;
        } else if (m_HubFilterMap.contains(HubFilterObject->m_sFilterName)) {
            QMessageBox::information(
                    this,
                    tr("Add hub filter"),
                    tr("A filter with that name already exists")
                    );

            delete HubFilterObject;
        } else {
            m_HubFilterMap.insert(HubFilterObject->m_sFilterName, HubFilterObject);

            ComboBox_FILTER->addItem(HubFilterObject->m_sFilterName);
        }

        DCConfig::Instance()->SaveDCHubFilter(&m_HubFilterMap);

        delete hf;
    }
}

/** */
void DCHubListManager::slotEditFilter() {
    QString s;
    DC_HubFilterObject * HubFilterObject;
    DCHubFilter * hf;

    s = ComboBox_FILTER->currentText();

    if ((!(s.isEmpty())) && (s != "Default")) {
        HubFilterObject = m_HubFilterMap.value(s);
        if (HubFilterObject != 0) {
            hf = new DCHubFilter(this);

            hf->Edit(HubFilterObject);

            if (hf->exec() == QDialog::Accepted) {
                // get the new data
                hf->Get(HubFilterObject);

                if (HubFilterObject->m_sFilterName.isEmpty()) {
                    HubFilterObject->m_sFilterName = s;

                    // todo: error message
                } else if (s != HubFilterObject->m_sFilterName) {
                    if (m_HubFilterMap.contains(HubFilterObject->m_sFilterName)) {
                        QMessageBox::information(
                                this,
                                tr("Edit hub filter"),
                                tr("A filter with that name already exists")
                                );
                        HubFilterObject->m_sFilterName = s;
                    } else {
                        /* not deleted because it's the same pointer */
                        m_HubFilterMap.remove(s);
                        m_HubFilterMap.insert(HubFilterObject->m_sFilterName, HubFilterObject);

                        ComboBox_FILTER->setItemText(ComboBox_FILTER->currentIndex(), HubFilterObject->m_sFilterName);
                    }
                }

                // update view
                slotFilterChange(HubFilterObject->m_sFilterName);

                DCConfig::Instance()->SaveDCHubFilter(&m_HubFilterMap);
            }

            delete hf;
        }
    }
}

/** */
void DCHubListManager::slotDelFilter() {
    QString s = ComboBox_FILTER->currentText();

    if ((!(s.isEmpty())) && (s != "Default")) {
        ComboBox_FILTER->removeItem(ComboBox_FILTER->currentIndex());

        QMap<QString, DC_HubFilterObject*>::const_iterator it = m_HubFilterMap.find(s);
        if (it != m_HubFilterMap.constEnd()) {
            delete it.value();
            m_HubFilterMap.remove(s);
            DCConfig::Instance()->SaveDCHubFilter(&m_HubFilterMap);
        } else {
            QMessageBox::warning(
                    this,
                    tr("Delete hub filter"),
                    tr("Filter not found")
                    );
        }
    }
}

/** */
void DCHubListManager::slotFilterChange(const QString & s) {
    if (!s.isEmpty()) {
        DC_HubFilterObject * filter = m_HubFilterMap.value(s);
        if (filter) {
            proxy->setFilter(filter);
            TabWidget_SERVERLIST->setTabText(0, tr("Public") + " (" + QString().setNum(TreeView_PUBLIC->model()->rowCount()) + "/" + QString().setNum(DCConfig::Instance()->GetPublicHubListSize()) + ")");
        }
    }
}

/** */
void DCHubListManager::RemoveBookmark(const QModelIndex &item) {
    if (!item.isValid())
        return;

    HubListModelItem *el = (HubListModelItem*)item.internalPointer();

    if (!el)
        return;

   if (DCConfig::Instance()->RemoveBookmarkHub(el->data(COLUMN_HUB_NAME).toString().toAscii().constData(),
                                               el->data(COLUMN_HUB_ADDRESS).toString().toAscii().constData(),
                                               el->data(COLUMN_HUB_DESC).toString().toAscii().constData())) {
        DCConfig::Instance()->SaveDCBookHub();

        // update tabbar
        TabWidget_SERVERLIST->setTabText(1, tr("Bookmarks") + " (" + QString().setNum(treeView_BOOKMARKS->model()->rowCount()) + ")");
    }
}

/** */
void DCHubListManager::AddBookmark(QString hubname, QString hubhost, QString description, bool autoconnect) {
    if (DCConfig::Instance()->AddBookmarkHub(hubname.toAscii().constData(), hubhost.toAscii().constData(), description.toAscii().constData())) {
        DCConfig::Instance()->SaveDCBookHub();

        DCConfigHubProfile hubprofile;

        //Loading defaults into the profile
        hubprofile.m_sNick = "";
        hubprofile.m_sPassword = "";
        hubprofile.m_sRemoteEncoding = "";
        hubprofile.m_bAutoConnect = autoconnect;
        hubprofile.m_sName = hubname.toAscii().constData();
        hubprofile.m_bSSL           = false;
        hubprofile.m_bTag           = false;
        hubprofile.m_bExtHubCount   = true;

        QList<QVariant> data;

        data << autoconnect << hubname << hubhost << tr("Default nick") << QString("") << description << tr("Default encoding");

        m_pHubListModel->addResult(data);

        // add new profile
        DCConfig::Instance()->AddHubProfile(&hubprofile);

        // save profile
        DCConfig::Instance()->SaveHubProfile();

        // set bookmark for that hub
        DCConfig::Instance()->SetBookmarkHubProfile(hubprofile.m_sName, hubprofile.m_sName);

        // update myinfo
        DCConnectionManager::Instance()->SendMyInfoToConnectedServers();

        // update tabbar
        TabWidget_SERVERLIST->setTabText(1, tr("Bookmarks") + " (" + QString().setNum(treeView_BOOKMARKS->model()->rowCount()) + ")");
    }

}

/** */
void DCHubListManager::UpdateBookmark(QString hubname, QString hubhost, QString description) {
   if (DCConfig::Instance()->UpdateBookmarkHub(hubname.toAscii().constData(), hubhost.toAscii().constData(), description.toAscii().constData())) {
        DCConfig::Instance()->SaveDCBookHub();
    }
}

/** */
CStringList<CString> * DCHubListManager::GetFilteredHubList() {
    CStringList<CString> * list = new CStringList<CString > ();

    for (int i = 0; i < proxy->rowCount(); ++i) {

        CString addr = proxy->data(proxy->index(i, 1)).toString().toAscii().constData();
        CString * string = 0;

        if (list->Get(addr, &string) != 0) {
            list->Add(addr, new CString(addr));
        }
    }

    if (list->Count() == 0) {
        delete list;
        list = 0;
    }

    return list;
}

/** */
void DCHubListManager::slotContextMenuBookmarkList(const QPoint &) {
    QModelIndexList mindexes = treeView_BOOKMARKS->selectionModel()->selectedIndexes();

    if (mindexes.size() == 0){
        return;
    }

    //HubListModelItem *item = (HubListModelItem*)mindexes.at(0).internalPointer();
    const QModelIndex &index = mindexes.at(0);

    if (!index.isValid())
        return;

    QMenu * m = new QMenu(this);

    QAction * connect = DCMenuHandler::addAction(m, emiCONNECT, (mindexes.count() > 0));
    DCMenuHandler::addAction(m, emiSEPARATOR);
    QAction * add = DCMenuHandler::addAction(m, emiADD);
    QAction * edit = DCMenuHandler::addAction(m, emiEDIT, true);
    QAction * remove = DCMenuHandler::addAction(m, emiREMOVE, (mindexes.count() > 0));
    DCMenuHandler::addAction(m, emiSEPARATOR);
    QAction * updserver = DCMenuHandler::addAction(m, emiUPDATE_SERVER, (mindexes.count() > 0));
    QAction * updall = DCMenuHandler::addAction(m, emiUPDATE_ALL_SERVER);
    DCMenuHandler::addAction(m, emiSEPARATOR);

    QAction * chosen = m->exec(QCursor::pos());

    delete m;

    if (chosen == 0) {
        return;
    }

    if (chosen == add) {
        slotAddBookmark();
    } else if (chosen == edit) {
        slotEditBookmark();
    } else if (chosen == remove) {
        if (queryRemoveBookmarks() == 1) {
            return;
        }

        DCConfigHubProfile hubprofile;
        for (int i = 0; i < mindexes.size(); i++) {
            const QModelIndex &item = mindexes.at(i);
            const HubListModelItem *el = (HubListModelItem*)item.internalPointer();

            // remove profile for that entry
            if (DCConfig::Instance()->GetHubProfile(el->data(COLUMN_HUB_NAME).toString().toAscii().constData(), &hubprofile)) {
                DCConfig::Instance()->DelHubProfile(hubprofile.m_sName);
            }
            // remove bookmark
            RemoveBookmark(mindexes.at(i));
        }

        // save changes to profile list, some may have been deleted
        DCConfig::Instance()->SaveHubProfile();

        // update myinfo
        DCConnectionManager::Instance()->SendMyInfoToConnectedServers();

        m_pHubListModel->clearModel();

        ShowBookmarkList();//Reload list
    } else if (chosen == connect) {
        for (int i = 0; i < mindexes.size(); i++) {
            const QModelIndex &item = mindexes.at(i);
            const HubListModelItem *el = (HubListModelItem*)item.internalPointer();

            DCConnectionManager::Instance()->Connect(el->data(COLUMN_HUB_NAME).toString().toAscii().constData(),
                                                     el->data(COLUMN_HUB_ADDRESS).toString().toAscii().constData());
        }
    } else if (chosen == updserver) {
        DCConfigHubItem hubitem;

        for (int i = 0; i < mindexes.size(); i++) {
            const QModelIndex &item = mindexes.at(i);
            HubListModelItem *el = (HubListModelItem*)item.internalPointer();

            if (DCConfig::Instance()->GetPublicHub(el->data(COLUMN_HUB_NAME).toString().toAscii().constData(), &hubitem)) {
                UpdateBookmark(hubitem.m_sName.Data(), hubitem.m_sHost.Data(), hubitem.m_sDescription.Data());
                el->updateColumn(COLUMN_HUB_ADDRESS, QVariant(QString(hubitem.m_sHost.Data())));
                el->updateColumn(COLUMN_HUB_DESC,    QVariant(QString(hubitem.m_sDescription.Data())));
            }
        }
    } else if (chosen == updall) {
        DCConfigHubItem * hubitem = 0;
        DCConfigHubItem hubitem1;
        CList<DCConfigHubItem> list;

        DCConfig::Instance()->GetBookmarkHubList(&list);

        while ((hubitem = list.Next(hubitem)) != 0) {
            if (DCConfig::Instance()->GetPublicHub(hubitem->m_sName, &hubitem1)) {
                DCConfig::Instance()->UpdateBookmarkHub(hubitem1.m_sName, hubitem1.m_sHost, hubitem1.m_sDescription);
            }
        }

        DCConfig::Instance()->SaveDCBookHub();

        m_pHubListModel->clearModel();

        ShowBookmarkList();
    }
}

/** */
void DCHubListManager::slotContextMenuPublicList(const QPoint & pos) {
    QList<PublicHubItem*> selitems;
    PublicHubItem * curitem = 0;
    QModelIndexList indexlist = TreeView_PUBLIC->selectionModel()->selectedIndexes();

    for (QModelIndexList::const_iterator it = indexlist.constBegin(); it != indexlist.constEnd(); ++it) {
        curitem = model->getItem(proxy->mapToSource(*it));

        if (curitem && (selitems.contains(curitem) == false)) {
            selitems.append(curitem);
        }
    }

    int numSelected = selitems.size();

    if (numSelected == 0) {
        return;
    }

    QMenu * m = new QMenu(this);

    QAction * connect = DCMenuHandler::addAction(m, emiCONNECT);
    DCMenuHandler::addAction(m, emiSEPARATOR);

    QAction * remove = 0;
    if (DCConfig::Instance()->GetRemovePublicHubEnabled()) {
        remove = DCMenuHandler::addAction(m, emiREMOVE);
        DCMenuHandler::addAction(m, emiSEPARATOR);
    }

    QAction * bookmark = DCMenuHandler::addAction(m, emiADD_BOOKMARK);
    DCMenuHandler::addAction(m, emiSEPARATOR);
    QAction * copycol = DCMenuHandler::addAction(m, emiCOPY_COLUMN_TO_CLIPBOARD, (numSelected == 1));
    QAction * copyrow = DCMenuHandler::addAction(m, emiCOPY_ROW_TO_CLIPBOARD, (numSelected == 1));

    QAction * chosen = m->exec(QCursor::pos());

    delete m;

    if (chosen == 0) {
        return;
    }

    if (chosen == connect) {
        for (QList<PublicHubItem*>::const_iterator it = selitems.constBegin(); it != selitems.constEnd(); ++it) {
            curitem = *it;
            DCConnectionManager::Instance()->Connect(curitem->name.toAscii().constData(), curitem->server.toAscii().constData());
        }
    } else if (chosen == remove) {
        for (QList<PublicHubItem*>::const_iterator it = selitems.constBegin(); it != selitems.constEnd(); ++it) {
            curitem = *it;

            if (DCConfig::Instance()->RemovePublicHub(curitem->name.toAscii().constData(), curitem->server.toAscii().constData(), curitem->description.toAscii().constData())) {
                model->deleteHub(curitem); // don't really need to check this return value too
            } else {
                QMessageBox::warning(
                        this, tr("Remove public hub"),
                        tr("Error removing hub \"") + curitem->name + "\" (" + curitem->server + ")"
                        );
            }
        }

        proxy->invalidate();
        slotFilterChange(ComboBox_FILTER->currentText());

        TabWidget_SERVERLIST->setTabText(0, tr("Public") + " (" + QString().setNum(TreeView_PUBLIC->model()->rowCount()) + "/" + QString().setNum(DCConfig::Instance()->GetPublicHubListSize()) + ")");
    } else if (chosen == bookmark) {
        for (QList<PublicHubItem*>::const_iterator it = selitems.constBegin(); it != selitems.constEnd(); ++it) {
            curitem = *it;
            AddBookmark(curitem->name, curitem->server, curitem->description);
        }

        TabWidget_SERVERLIST->setTabText(1, tr("Bookmarks") + " (" + QString().setNum(treeView_BOOKMARKS->model()->rowCount()) + ")");
    } else if (chosen == copycol) {
        int column = TreeView_PUBLIC->columnAt(pos.x());

        curitem = selitems.first();

        QString s;

        switch (column) {
            case 0: s = curitem->name;
                break;
            case 1: s = curitem->server;
                break;
            case 2: s = curitem->description;
                break;
            case 3: s = QString().setNum(curitem->users);
                break;
            case 4: s = curitem->country;
                break;
            case 5: s = DCGuiUtils::GetSizeString(curitem->shared);
                break;
            case 6: s = DCGuiUtils::GetSizeString(curitem->minshare);
                break;
            case 7: s = curitem->extra;
                break;
            default: break;
        }

        QApplication::clipboard()->setText(s);
    } else if (chosen == copyrow) {
        QString s;

        curitem = selitems.first();

        s = curitem->name + " " + curitem->server + " "
                + curitem->description + " " + QString().setNum(curitem->users)
                + " " + curitem->country + " " + DCGuiUtils::GetSizeString(curitem->shared)
                + " " + DCGuiUtils::GetSizeString(curitem->minshare)
                + " " + curitem->extra;

        QApplication::clipboard()->setText(s);
    }
}

/** update CConfig's sorted bookmarks map */
void DCHubListManager::updateBookmarkOrder() {
    if (!(treeView_BOOKMARKS->isSortingEnabled())) {
        //printf("updateBookmarkOrder cancelled, widget is not sorted\n");
        return;
    }

    if (!m_pHubListModel)
        return;

    const int count = m_pHubListModel->rowCount(QModelIndex());
    std::vector<int> newpositions(count);

    int i = 0, j = 0;
    const QList<HubListModelItem*> &items = m_pHubListModel->getItems();

    for (; j < items.size(); j++) {

        HubListModelItem *item = items.at(j);

        //printf("%s %d->%d\n",item->text(0).toAscii().constData(),item->position,i);
        if ((item->position < 0) || (item->position >= count)) {
            printf("updateBookmarkOrder: Existing key for %s out of range at %d\n", item->data(COLUMN_HUB_NAME).toString().toAscii().constData(),
                   item->position);
            return;
        }
        newpositions[item->position] = i;
        item->position = i;

        ++i;
    }

    //printf("call ReorderBookmarkHubs\n");

    DCConfig::Instance()->ReorderBookmarkHubs(newpositions);
    DCConfig::Instance()->SaveDCBookHub();

}

/** */
void DCHubListManager::slotMoveBookmarkUp() {
    moveBookmark(false);
}

/** */
void DCHubListManager::slotMoveBookmarkDown() {
    moveBookmark(true);
}

void DCHubListManager::moveBookmark(bool down){
    QModelIndexList mindexes = treeView_BOOKMARKS->selectionModel()->selectedIndexes();

    if (mindexes.size() == 0){
        return;
    }

    HubListModelItem *item = (HubListModelItem*)mindexes.at(0).internalPointer();
    const QModelIndex &index = mindexes.at(0);

    if (!index.isValid())
        return;

    int old_index = m_pHubListModel->getItems().indexOf(item), new_index;

    if (down)
        m_pHubListModel->moveDown(index);
    else
        m_pHubListModel->moveUp(index);

    new_index = m_pHubListModel->getItems().indexOf(item);

    //printf("Reorder item from %i to %i\n", old_index, new_index);

    //updateBookmarkOrder();

    DCConfig::Instance()->MoveBookmarkHub(old_index, new_index);
    DCConfig::Instance()->SaveDCBookHub();

    m_pHubListModel->repaint();

    treeView_BOOKMARKS->selectionModel()->setCurrentIndex(index,
                                                          QItemSelectionModel::Clear);

    for (int i = COLUMN_HUB_AUTOCONNECT; i <= COLUMN_HUB_ENCODING; i++)
        treeView_BOOKMARKS->selectionModel()->setCurrentIndex(m_pHubListModel->index(item->row(), i, QModelIndex()),
                                                              QItemSelectionModel::Select);
}

/** */
void DCHubListManager::slotSaveBookmarks() {
    /*
     * Of couse updateBookmarkOrder() could just have been made a slot, but due to the large
     * number of problems encountered while attempting to have the code in updateBookmarkOrder
     * called automatically after the QTreeWidget has been sorted ( which appears to be impossible -
     * you can connect to QAbstractItemModel::layoutChanged() or various QHeaderView signals
     * but these are called either before the QTreeWidget has been sorted, or worse, during )
     * so it is not a slot to indicate that it is not called automatically and updateBookmarkOrder()
     * must be called manually at various points and on shutdown.
     */
    if (treeView_BOOKMARKS->isSortingEnabled()) {
        updateBookmarkOrder();
    } else {
        DCConfig::Instance()->SaveDCBookHub();
    }
}

/** */
void DCHubListManager::slotAltRowColors(bool enable) {
    treeView_BOOKMARKS->setAlternatingRowColors(enable);
    TreeView_PUBLIC->setAlternatingRowColors(enable);
}

/** */
void DCHubListManager::slotAddBookmark() {
    DCEditServer * des = new DCEditServer(this);
    des->setWindowTitle(tr("Add Bookmark"));

    //reset values
    des->Reset();

    // DCEditServer now checks name and host are not empty
    if (des->exec() == QDialog::Accepted) {
        QString hub_name, hub_addr, hub_nick, hub_pass, hub_desc, hub_enc;
        bool hub_auto = false;
        QList<QVariant> data;

        // get profile from dialog
        DCConfigHubProfile hubprofile;
        des->GetProfile(&hubprofile);

        hub_name = des->LineEdit_NAME->text();
        hub_addr = des->LineEdit_HOST->text();
        hub_desc = des->LineEdit_DESCRIPTION->text();
        hub_nick = QString(hubprofile.m_sNick.Data());
        hub_pass = QString(hubprofile.m_sPassword.Data());
        hub_enc  = QString(hubprofile.m_sRemoteEncoding.Data());
        hub_auto = hubprofile.m_bAutoConnect;

        if (hub_nick.isEmpty() || hub_nick.isNull())
            hub_nick = tr("Default nick");
        if (hub_enc.isEmpty()  || hub_enc.isNull())
            hub_enc  = tr("Default encoding");

        data << hub_auto << hub_name << hub_addr << hub_nick << hub_pass << hub_desc << hub_enc;

        if (DCConfig::Instance()->AddBookmarkHub(hub_name.toAscii().constData(), hub_addr.toAscii().constData(), hub_desc.toAscii().constData())){
            m_pHubListModel->addResult(data);

            DCConfig::Instance()->SaveDCBookHub();
        }
        // add new profile
        DCConfig::Instance()->AddHubProfile(&hubprofile);

        // save profile
        DCConfig::Instance()->SaveHubProfile();

        // set bookmark for that hub
        DCConfig::Instance()->SetBookmarkHubProfile(des->LineEdit_NAME->text().toAscii().constData(), des->LineEdit_NAME->text().toAscii().constData());
        // update myinfo
        DCConnectionManager::Instance()->SendMyInfoToConnectedServers();
    }

    delete des;
}

/** */
void DCHubListManager::slotEditBookmark() {
    QModelIndexList mindexes = treeView_BOOKMARKS->selectionModel()->selectedIndexes();

    if (mindexes.size() == 0){
        return;
    }

    HubListModelItem *item = (HubListModelItem*)mindexes.at(0).internalPointer();
    const QModelIndex &index = mindexes.at(0);

    if (!index.isValid())
        return;

    DCConfigHubProfile hubprofile;
    DCConfigHubItem ConfigHubItem;

    DCEditServer * des = new DCEditServer(this);
    des->setWindowTitle(tr("Edit Bookmark"));

    QString hub_name = item->data(COLUMN_HUB_NAME).toString();

    des->LineEdit_NAME->setText(hub_name);
    des->LineEdit_HOST->setText(item->data(COLUMN_HUB_ADDRESS).toString());
    des->LineEdit_DESCRIPTION->setText(item->data(COLUMN_HUB_DESC).toString());
    des->CheckBox_AUTOCONNECT->setChecked(item->data(COLUMN_HUB_AUTOCONNECT).toBool());

    // hubitem
   if (DCConfig::Instance()->GetBookmarkHub(hub_name.toAscii().constData(), &ConfigHubItem)) {
        if (ConfigHubItem.m_sProfile.NotEmpty()) {
            // get profile
            if (DCConfig::Instance()->GetHubProfile(ConfigHubItem.m_sProfile, &hubprofile)) {
                des->SetProfile(&hubprofile);
            }
        }
    }

    // DCEditServer now checks that name and host are not empty
    if (des->exec() == QDialog::Accepted) {
        //RemoveBookmark(curitem);

        // remove the profile
        DCConfig::Instance()->DelHubProfile(hubprofile.m_sName);

        // add as new bookmark
        /*AddBookmark(des->LineEdit_NAME->text(), des->LineEdit_HOST->text(),
                    des->LineEdit_DESCRIPTION->text(), des->CheckBox_AUTOCONNECT->isChecked());*/

        // check if profiling possible
        des->GetProfile(&hubprofile);

        // add new profile
        DCConfig::Instance()->AddHubProfile(&hubprofile);

        // save profile
        DCConfig::Instance()->SaveHubProfile();

        // set bookmark for that hub
        DCConfig::Instance()->SetBookmarkHubProfile(des->LineEdit_NAME->text().toAscii().constData(), des->LineEdit_NAME->text().toAscii().constData());

        // update myinfo
        DCConnectionManager::Instance()->SendMyInfoToConnectedServers();

        QString hub_name, hub_addr, hub_nick, hub_pass, hub_desc, hub_enc;
        bool hub_auto = false;

        hub_name = des->LineEdit_NAME->text();
        hub_addr = des->LineEdit_HOST->text();
        hub_desc = des->LineEdit_DESCRIPTION->text();
        hub_nick = QString(hubprofile.m_sNick.Data());
        hub_pass = QString(hubprofile.m_sPassword.Data());
        hub_enc  = QString(hubprofile.m_sRemoteEncoding.Data());
        hub_auto = hubprofile.m_bAutoConnect;

        //Update ModelIndex
        item->updateColumn(COLUMN_HUB_ADDRESS, hub_addr);
        item->updateColumn(COLUMN_HUB_AUTOCONNECT, hub_auto);
        item->updateColumn(COLUMN_HUB_DESC, hub_desc);
        item->updateColumn(COLUMN_HUB_ENCODING, hub_enc);
        item->updateColumn(COLUMN_HUB_NAME, hub_name);
        item->updateColumn(COLUMN_HUB_NICK, hub_nick);
        item->updateColumn(COLUMN_HUB_PASSWORD, hub_pass);

        m_pHubListModel->repaint();
    }

    delete des;
}

/** */
void DCHubListManager::slotDelBookmark() {
    QModelIndexList mindexes = treeView_BOOKMARKS->selectionModel()->selectedIndexes();

    if (mindexes.size() == 0){
        return;
    }

    HubListModelItem *item = (HubListModelItem*)mindexes.at(0).internalPointer();
    const QModelIndex &index = mindexes.at(0);

    if (!index.isValid() || !item)
        return;

    if (queryRemoveBookmarks() == 1) {
        return;
    }

    // remove profile for that entry
    DCConfigHubProfile hubprofile;
    if (DCConfig::Instance()->GetHubProfile(item->data(COLUMN_HUB_NAME).toString().toAscii().constData(), &hubprofile)) {
        DCConfig::Instance()->DelHubProfile(hubprofile.m_sName);
    }

    // remove bookmark
    RemoveBookmark(index);

    // save changes to profile list, some may have been deleted
    DCConfig::Instance()->SaveHubProfile();

    // update myinfo
    DCConnectionManager::Instance()->SendMyInfoToConnectedServers();

    ShowBookmarkList();
}

int DCHubListManager::queryRemoveBookmarks() {
    StringMap * map;
    int i;

    if (DCConfig::Instance()->GetMap("DIALOG_REMOVE_BOOKMARK", map) == false) {
        (*map)["SHOW"] = QString::number(1);
        (*map)["DEFAULT"] = QString::number(0);
    }

    if (((*map)["SHOW"].toInt() == 1) || ((*map)["DEFAULT"].toInt() == 1)) {
        CDialogMessage * dlg = new CDialogMessage(
                this,
                QMessageBox::Warning,
                tr("Remove bookmarks"),
                tr("You are sure ?"),
                tr("Remove"),
                tr("Cancel")
                );

        i = dlg->exec();

        (*map)["SHOW"] = QString::number(!dlg->GetCheckBoxStatus());
        (*map)["DEFAULT"] = QString::number(i);

        delete dlg;
    } else {
        i = (*map)["DEFAULT"].toInt();
    }

    return i;
}

void DCHubListManager::slotAutoConnectToAllHubs() {
    DCConfig::Instance()->SetAutoConnectToAllHubs(checkBox_AUTOCONNECT->isChecked());
}
