/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "filebrowsermodel.h"

#include <QtGui>
#include <QFileInfo>
#include <QList>
#include <QStringList>
#include <QPalette>
#include <QColor>
#include <QIcon>
#include <QPixmap>
#include <QFontMetrics>
#include <QSize>

#include <dclib/cfilemanager.h>
#include "dciconloader.h"
#include "dcguiutils.h"
#include "dcgui.h"

#include <dclib/cfilemanager.h>
#include <set>

FileBrowserModel::FileBrowserModel(QObject *parent)
    : QAbstractItemModel(parent), iconsScaled(false)
{
    rootItem = new FileBrowserItem(QList<QVariant>());

    sortColumn = COLUMN_FILEBROWSER_TYPE;
    sortOrder = Qt::DescendingOrder;
}

FileBrowserModel::~FileBrowserModel()
{
    if (rootItem)
        delete rootItem;
}

int FileBrowserModel::columnCount(const QModelIndex &parent) const
{
    if (parent.isValid())
        return static_cast<FileBrowserItem*>(parent.internalPointer())->columnCount();
    else
        return rootItem->columnCount();
}

QVariant FileBrowserModel::data(const QModelIndex &index, int role) const
{
    if (!index.isValid())
        return QVariant();

    FileBrowserItem *item = static_cast<FileBrowserItem*>(index.internalPointer());

    switch(role) {
        case Qt::DecorationRole:
        {
            if (index.column() != COLUMN_FILEBROWSER_NAME)
                break;

            QPixmap icon;

            if (item->isDir){
                icon = DCIconLoader::Instance()->GetPixmap(eiFOLDER_BLUE);

                if (iconsScaled)
                    icon = icon.scaled(iconsSize);

                return icon;
            }

            switch (item->type) {
                case eftMP3:
                    icon = DCIconLoader::Instance()->GetPixmap(eiFILETYPE_MP3);
                    break;
                case eftARCHIVE:
                    icon = DCIconLoader::Instance()->GetPixmap(eiFILETYPE_ARCHIVE);
                    break;
                case eftDOCUMENT:
                    icon = DCIconLoader::Instance()->GetPixmap(eiFILETYPE_DOCUMENT);
                    break;
                case eftAPPLICATION:
                    icon = DCIconLoader::Instance()->GetPixmap(eiFILETYPE_APPLICATION);
                    break;
                case eftPICTURE:
                    icon = DCIconLoader::Instance()->GetPixmap(eiFILETYPE_PICTURE);
                    break;
                case eftVIDEO:
                    icon = DCIconLoader::Instance()->GetPixmap(eiFILETYPE_VIDEO);
                    break;
                default:
                    icon = DCIconLoader::Instance()->GetPixmap(eiFILETYPE_UNKNOWN);
                    break;
            }

            if (iconsScaled)
                icon = icon.scaled(iconsSize);

            return icon;
        }
        case Qt::DisplayRole:
        {
            return item->data(index.column());
        }
        case Qt::TextAlignmentRole:
        {
            bool align_right = (index.column() == COLUMN_FILEBROWSER_ESIZE) || (index.column() == COLUMN_FILEBROWSER_SIZE);

            if (align_right)
                return Qt::AlignRight;
            else
                return Qt::AlignLeft;
        }
        case Qt::ForegroundRole:
        {
            if (item->isDir)
                break;

            if (!item->cfile_flag){
                std::set<unsigned long> *my_set = CFileManager::Instance()->SearchHash(
                                                                                      item->data(COLUMN_FILEBROWSER_TTH).toString().toUtf8().constData()
                                                                                      );

                item->cfile_flag = true;

                if (my_set && my_set->size() > 0){
                    item->highlight = true;

                    delete my_set;
                }
            }

            if (item->highlight)
                return QColor(0x00, 0x80, 0x80);

            break;
        }
        case Qt::BackgroundColorRole:
            break;
        case Qt::ToolTipRole:
        {
            QString tooltip = "<b>"+item->data(COLUMN_FILEBROWSER_NAME).toString()+"</b><br/>";

            if (item->parent() && item->parent()->parent())
                tooltip += QString("<b>Path: </b>") + createRemotePath(item->parent()) + "<br/>";

            tooltip += tr("Type: ");

            if (item->isDir){
                tooltip += tr("<b>Directory</b><br/>");
                tooltip += tr("Total size: <b>")
                           +DCGuiUtils::GetSizeString(item->data(COLUMN_FILEBROWSER_ESIZE).toULongLong())
                           +" ("
                           +QString().setNum(item->data(COLUMN_FILEBROWSER_ESIZE).toULongLong())
                           +") </b><br/>";
                if (item->childCount() > 0){
                    int i = 0;

                    tooltip += tr("Child items: <br/>");

                    while ((i < item->childCount()) && (i < 5)){
                        const FileBrowserItem *child = item->childItems.at(i);

                        if (child->isDir)
                            tooltip += tr("  Directory: ");
                        else
                            tooltip += tr("  File: ");

                        tooltip += "<b>" + child->data(COLUMN_FILEBROWSER_NAME).toString()
                                   + "</b> - <b>"
                                   + DCGuiUtils::GetSizeString(child->data(COLUMN_FILEBROWSER_ESIZE).toULongLong())
                                   + "</b><br/>";

                        i++;
                    }
                    
                }
                
                if (item->childCount() > 4)
                    tooltip += "......<br/>";
            }
            else{
                tooltip += tr("<b>File</b><br/>");
                tooltip += tr("Total size: <b>")
                           +DCGuiUtils::GetSizeString(item->data(COLUMN_FILEBROWSER_ESIZE).toULongLong())
                           +" ("
                           +QString().setNum(item->data(COLUMN_FILEBROWSER_ESIZE).toULongLong())
                           +") </b><br/>";
            }

            return tooltip;
        }
    }

    return QVariant();
}

namespace {

template <Qt::SortOrder order>
struct Compare {
    void static sort(int col, QList<FileBrowserItem*>& items) {
        qStableSort(items.begin(), items.end(), getAttrComp(col));
    }

    void static insertSorted(int col, QList<FileBrowserItem*>& items, FileBrowserItem* item) {
        QList<FileBrowserItem*>::iterator it = qLowerBound(items.begin(), items.end(), item, getAttrComp(col));
        items.insert(it, item);
    }

    private:
        typedef bool (*AttrComp)(const FileBrowserItem * l, const FileBrowserItem * r);
        AttrComp static getAttrComp(int column) {
            switch (column){
                case COLUMN_FILEBROWSER_NAME:
                    return AttrCmp<COLUMN_FILEBROWSER_NAME>;
                case COLUMN_FILEBROWSER_ESIZE:
                    return NumCmp<COLUMN_FILEBROWSER_ESIZE>;
                case COLUMN_FILEBROWSER_SIZE:
                    return NumCmp<COLUMN_FILEBROWSER_ESIZE>;
                case COLUMN_FILEBROWSER_TTH:
                    return AttrCmp<COLUMN_FILEBROWSER_TTH>;
                case COLUMN_FILEBROWSER_TYPE:
                    return AttrCmp<COLUMN_FILEBROWSER_TYPE>;
            }

            Q_ASSERT_X(false, "getAttrComp", QString("Inncorrect column %1").arg(column).toAscii().constData());
            return 0;
        }
        template <int i>
        bool static AttrCmp(const FileBrowserItem * l, const FileBrowserItem * r) {
            if (l->isDir != r->isDir && i != COLUMN_FILEBROWSER_TYPE){
                return l->isDir;
            }
            return Cmp(QString::localeAwareCompare(l->data(i).toString(), r->data(i).toString()), 0);
        }
        template <int column>
        bool static NumCmp(const FileBrowserItem * l, const FileBrowserItem * r) {
            if (l->isDir != r->isDir && column != COLUMN_FILEBROWSER_TYPE){
                return l->isDir;
            }
            return Cmp(l->data(column).toULongLong(), r->data(column).toULongLong());
       }
        template <typename T>
        bool static Cmp(const T& l, const T& r);
};

template <> template <typename T>
bool inline Compare<Qt::AscendingOrder>::Cmp(const T& l, const T& r) {
    return l < r;
}

template <> template <typename T>
bool inline Compare<Qt::DescendingOrder>::Cmp(const T& l, const T& r) {
    return l > r;
}
}

QVariant FileBrowserModel::headerData(int section, Qt::Orientation orientation,
                               int role) const
{
    QList<QVariant> rootData;
    rootData << tr("Name") << tr("Size") << tr("Exact size") << tr("Type") << tr("TTH");

    if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
        return rootData.at(section);

    return QVariant();
}

QModelIndex FileBrowserModel::index(int row, int column, const QModelIndex &parent)
            const
{
    if (!hasIndex(row, column, parent))
        return QModelIndex();

    FileBrowserItem *parentItem;

    if (!parent.isValid())
        parentItem = rootItem;
    else
        parentItem = static_cast<FileBrowserItem*>(parent.internalPointer());

    FileBrowserItem *childItem = parentItem->child(row);
    if (childItem)
        return createIndex(row, column, childItem);
    else
        return QModelIndex();
}

QModelIndex FileBrowserModel::parent(const QModelIndex &index) const
{
    if (!index.isValid())
        return QModelIndex();

    FileBrowserItem *childItem = static_cast<FileBrowserItem*>(index.internalPointer());
    FileBrowserItem *parentItem = childItem->parent();

    if (parentItem == rootItem)
        return QModelIndex();

    return createIndex(parentItem->row(), 0, parentItem);
}

int FileBrowserModel::rowCount(const QModelIndex &parent) const
{
    FileBrowserItem *parentItem;
    if (parent.column() > 0)
        return 0;

    if (!parent.isValid())
        parentItem = rootItem;
    else
        parentItem = static_cast<FileBrowserItem*>(parent.internalPointer());

    return parentItem->childCount();
}

void FileBrowserModel::sort(int column, Qt::SortOrder order) {
    sortColumn = column;
    sortOrder = order;

    if (!rootItem || rootItem->childItems.empty())
        return;

    if (column == -1)
        return;

    emit layoutAboutToBeChanged();

    if (order == Qt::AscendingOrder)
        Compare<Qt::AscendingOrder>().sort(column, rootItem->childItems);
    else if (order == Qt::DescendingOrder)
        Compare<Qt::DescendingOrder>().sort(column, rootItem->childItems);

    emit layoutChanged();
}

void FileBrowserModel::addResult
        (
        const QString &name,
        const QString &tth,
        const qulonglong size,
        bool isDir
        )
{
    FileBrowserItem *item;

    eFileTypes type = CFileManager::Instance()->GetFileType(name.toAscii().constData());

    item = new FileBrowserItem(QList<QVariant>()<< name << DCGuiUtils::GetSizeString(size) << size << tth,
                              rootItem);
    item->type = type;
    item->isDir = isDir;

    rootItem->appendChild(item);

    emit layoutChanged();
}

void FileBrowserModel::setRootElem(FileBrowserItem *root, bool del_old, bool controlNull){
    if (controlNull && !root)
        return;

    FileBrowserItem *from = rootItem, *to = root;

    if (del_old && rootItem){
        delete rootItem;

        rootItem = NULL;
    }

    rootItem = root;

    if (rootItem){
        emit layoutChanged();
    }

    emit rootChanged(from, to);
}

FileBrowserItem *FileBrowserModel::getRootElem() const{
    return rootItem;
}

void FileBrowserModel::setIconsScaled(bool scaled, const QSize &size){
    iconsScaled = scaled;
    iconsSize = size;
}

int FileBrowserModel::getSortColumn() const {
    return sortColumn;
}

void FileBrowserModel::setSortColumn(int c) {
    sortColumn = c;
}

Qt::SortOrder FileBrowserModel::getSortOrder() const {
    return sortOrder;
}

void FileBrowserModel::setSortOrder(Qt::SortOrder o) {
    sortOrder = o;
}

QString FileBrowserModel::createRemotePath(FileBrowserItem *item) const{
    QString s;
    FileBrowserItem * pitem;

    if (!item) {
        return s;
    }

    pitem = item;
    s = pitem->data(COLUMN_FILEBROWSER_NAME).toString();

    while ((pitem = pitem->parent()) != NULL) {
        // check for root entry
        if (pitem->parent() != NULL) {
            s = pitem->data(COLUMN_FILEBROWSER_NAME).toString() + "\\" + s;
        }
    }

    s.replace("\\\\\\\\", "\\");

    return s;
}

FileBrowserItem *FileBrowserModel::createRootForPath(const QString &path, FileBrowserItem *pathRoot){
    if (path.isEmpty() || path.isNull())
        return NULL;

    QString _path = path;
    _path.replace("\\", "/");

    QStringList list = _path.split("/");
    FileBrowserItem *root = pathRoot?pathRoot:rootItem;

    if (list.empty() || !root)
        return NULL;

    foreach (QString s, list){
        if (s.isEmpty() || s.isNull())
            continue;

        if (!root)
            return NULL;

        if (s == ".." && root->parent()){
            root = root->parent();

            continue;
        }
        else if (s == ".")
            continue;

        bool found = false;

        foreach(FileBrowserItem *item, root->childItems){
            if (!item->isDir)
                continue;

            QString name = item->data(COLUMN_FILEBROWSER_NAME).toString();

            if (name == s){
                root = item;
                found = true;

                break;
            }
        }

        if (!found)
            return root;

    }

    return root;
}

QModelIndex FileBrowserModel::createIndexForItem(FileBrowserItem *item){
    if (!rootItem || !item)
        return QModelIndex();

    QStack<FileBrowserItem*> stack;
    FileBrowserItem *root = item->parent();

    while (root && (root != rootItem)){
        stack.push(root);

        root = root->parent();
    }

    QModelIndex parent = QModelIndex();
    QModelIndex child;

    while (!stack.empty()){
        FileBrowserItem *el = stack.pop();

        parent = index(el->row(), COLUMN_FILEBROWSER_NAME, parent);
    }

    return index(item->row(), COLUMN_FILEBROWSER_NAME, parent);
}

FileBrowserItem::FileBrowserItem(const QList<QVariant> &data, FileBrowserItem *parent) :
    itemData(data), parentItem(parent), isDir(false), highlight(false), cfile_flag(false)
{
}

FileBrowserItem::FileBrowserItem(const FileBrowserItem &item){
    itemData = item.itemData;
    isDir = item.isDir;
    highlight = item.highlight;
    cfile_flag = item.cfile_flag;
    type = item.type;
}
void FileBrowserItem::operator=(const FileBrowserItem &item){
    itemData = item.itemData;
    isDir = item.isDir;
    highlight = item.highlight;
    cfile_flag = item.cfile_flag;
    type = item.type;
}

FileBrowserItem::~FileBrowserItem()
{
    if (childItems.size() > 0)
        qDeleteAll(childItems);
}

void FileBrowserItem::appendChild(FileBrowserItem *item) {
    childItems.append(item);
}

FileBrowserItem *FileBrowserItem::child(int row) {
    return childItems.value(row);
}

int FileBrowserItem::childCount() const {
    return childItems.count();
}

int FileBrowserItem::columnCount() const {
    return itemData.count();
}

QVariant FileBrowserItem::data(int column) const {
    return itemData.value(column);
}

FileBrowserItem *FileBrowserItem::parent() {
    return parentItem;
}

int FileBrowserItem::row() const {
    if (parentItem)
        return parentItem->childItems.indexOf(const_cast<FileBrowserItem*>(this));

    return 0;
}

void FileBrowserItem::updateColumn(int column, QVariant var){
    if (column > (itemData.size()-1))
        return;

    itemData[column] = var;
}

FileBrowserItem *FileBrowserItem::nextSibling(){
    if (!parent())
        return NULL;

    if (row() == (parent()->childCount()-1))
        return NULL;

    return parent()->child(row()+1);
}
