/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "hublistmodel.h"

#include <QList>
#include <QStringList>

HubListModel::HubListModel(QObject *parent)
    : QAbstractItemModel(parent)
{
    QList<QVariant> rootData;
    rootData << tr("Autoconnect") << tr("Name") << tr("Address")
             << tr("Nick") << tr("Password") << tr("Description")
             << tr("Remote encoding");

    rootItem = new HubListModelItem(rootData, NULL);
}

HubListModel::~HubListModel()
{
    delete rootItem;
}

int HubListModel::columnCount(const QModelIndex &parent) const
{
    if (parent.isValid())
        return static_cast<HubListModelItem*>(parent.internalPointer())->columnCount();
    else
        return rootItem->columnCount();
}

QVariant HubListModel::data(const QModelIndex &index, int role) const
{
    if (!index.isValid())
        return QVariant();

    if (index.column() > columnCount(QModelIndex()))
        return QVariant();

    HubListModelItem *item = static_cast<HubListModelItem*>(index.internalPointer());

    switch(role) {
        case Qt::DecorationRole: // icon
            break;
        case Qt::DisplayRole:
            if (index.column() == COLUMN_HUB_AUTOCONNECT)
                break;
            else if (index.column() != COLUMN_HUB_PASSWORD)
                return item->data(index.column());
            else
                return tr("******");

            break;
        case Qt::TextAlignmentRole:
            break;
        case Qt::ForegroundRole:
            break;
        case Qt::ToolTipRole:
            break;
        case Qt::CheckStateRole:
            if (index.column() == COLUMN_HUB_AUTOCONNECT)
                return item->data(COLUMN_HUB_AUTOCONNECT);
            break;
    }

    return QVariant();
}

Qt::ItemFlags HubListModel::flags(const QModelIndex &index) const
{
    if (!index.isValid())
        return 0;

    Qt::ItemFlags flags = Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsEditable;

    if (index.column() == COLUMN_HUB_AUTOCONNECT)
        flags |= Qt::ItemIsUserCheckable;
    else if (index.column() == COLUMN_HUB_PASSWORD || index.column() == COLUMN_HUB_ENCODING)
        flags &= ~Qt::ItemIsEditable;

    return flags;
}

QVariant HubListModel::headerData(int section, Qt::Orientation orientation,
                               int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
        return rootItem->data(section);

    return QVariant();
}

QModelIndex HubListModel::index(int row, int column, const QModelIndex &parent)
            const
{
    if (!hasIndex(row, column, parent))
        return QModelIndex();

    HubListModelItem *parentItem;

    if (!parent.isValid())
        parentItem = rootItem;
    else
        parentItem = static_cast<HubListModelItem*>(parent.internalPointer());

    HubListModelItem *childItem = parentItem->child(row);
    if (childItem)
        return createIndex(row, column, childItem);
    else
        return QModelIndex();
}

QModelIndex HubListModel::parent(const QModelIndex &index) const
{
    if (!index.isValid())
        return QModelIndex();

    HubListModelItem *childItem = static_cast<HubListModelItem*>(index.internalPointer());
    HubListModelItem *parentItem = childItem->parent();

    if (parentItem == rootItem)
        return QModelIndex();

    return createIndex(parentItem->row(), 0, parentItem);
}

int HubListModel::rowCount(const QModelIndex &parent) const
{
    HubListModelItem *parentItem;
    if (parent.column() > 0)
        return 0;

    if (!parent.isValid())
        parentItem = rootItem;
    else
        parentItem = static_cast<HubListModelItem*>(parent.internalPointer());

    return parentItem->childCount();
}

void HubListModel::sort(int column, Qt::SortOrder order) {
    return;
}

void HubListModel::clearModel(){
    QList<HubListModelItem*> childs = rootItem->childItems;   //Copying list in another place
    rootItem->childItems.clear();                             //Cleaning root of the model. Do not delete items directly from the root item

    qDeleteAll(childs);

    rootItem->childItems.clear();

    reset();

    emit layoutChanged();
}

bool HubListModel::removeItem(const QModelIndex &el){
    if (!el.isValid())
        return false;

    HubListModelItem *item = (HubListModelItem*)el.internalPointer();

    if (!item || !rootItem->childItems.contains(item))
        return false;

    beginRemoveRows(QModelIndex(), item->row(), item->row());
    rootItem->childItems.removeAt(rootItem->childItems.indexOf(item));
    endRemoveRows();

    delete item;

    emit layoutChanged();

    return true;
}

void HubListModel::moveUp(const QModelIndex &index){
    if (!index.isValid())
        return;

    HubListModelItem *el = (HubListModelItem*)index.internalPointer();

    if (!el || !rootItem->childItems.contains(el))
        return;

    if (rootItem->childItems.size() == 1)
        return;

    int old_index = rootItem->childItems.indexOf(el), new_index;
    HubListModelItem *old_el;

    if (old_index == 0)//already top item
        return;

    new_index = (old_index - 1);

    old_el = rootItem->childItems.at(new_index);

    rootItem->childItems[new_index] = el;
    rootItem->childItems[old_index] = old_el;
}

void HubListModel::moveDown(const QModelIndex &index){
    if (!index.isValid())
        return;

    HubListModelItem *el = (HubListModelItem*)index.internalPointer();

    if (!el || !rootItem->childItems.contains(el))
        return;

    if (rootItem->childItems.size() == 1)
        return;

    int old_index = rootItem->childItems.indexOf(el), new_index;
    HubListModelItem *old_el;

    if (old_index == (rootItem->childItems.size() - 1))//already bottom item
        return;

    new_index = (old_index + 1);

    old_el = rootItem->childItems.at(new_index);

    rootItem->childItems[new_index] = el;
    rootItem->childItems[old_index] = old_el;
}

void HubListModel::addResult(QList<QVariant> &data){

    HubListModelItem *item = new HubListModelItem(data, rootItem);
    rootItem->appendChild(item);

    item->position = item->row();

    emit layoutChanged();
}

const QList<HubListModelItem*> &HubListModel::getItems(){
    return rootItem->childItems;
}

void HubListModel::repaint(){
    emit layoutChanged();
}

HubListModelItem::HubListModelItem(const QList<QVariant> &data, HubListModelItem *parent) :
    itemData(data), parentItem(parent), position(-1)
{
}

HubListModelItem::~HubListModelItem()
{
    if (childItems.size() > 0)
        qDeleteAll(childItems);
}

void HubListModelItem::appendChild(HubListModelItem *item) {
    childItems.append(item);
}

HubListModelItem *HubListModelItem::child(int row) {
    return childItems.value(row);
}

int HubListModelItem::childCount() const {
    return childItems.count();
}

int HubListModelItem::columnCount() const {
    return itemData.count();
}

QVariant HubListModelItem::data(int column) const {
    return itemData.at(column);
}

HubListModelItem *HubListModelItem::parent() {
    return parentItem;
}

int HubListModelItem::row() const {
    if (parentItem)
        return parentItem->childItems.indexOf(const_cast<HubListModelItem*>(this));

    return 0;
}

void HubListModelItem::updateColumn(unsigned column, QVariant var){
    if (column > (itemData.size()-1))
        return;

    itemData[column] = var;
}
